@@ -4,12 +4,12 @@ import asyncio
 import json
 from collections import defaultdict
 from functools import wraps
-from typing import List, Tuple, Optional, Dict, Any, Union
+from typing import List, Tuple, Optional, Dict, Any, Union, Coroutine
 from concurrent.futures import ThreadPoolExecutor
 from datetime import datetime
 
 import aiohttp
-import pymysql
+import aiomysql as aiomysql
 from fuzzywuzzy import fuzz
 from pymysql import Connection, InterfaceError
 from pymysql.cursors import Cursor
@@ -92,13 +92,12 @@ def replace_aliases(name: str, aliases: dict, threshold=80) -> str:
     return ' '.join(new_words)
 
 
-def find_best_match(item_name: str, items: List[Dict[str, Any]]) -> Tuple[int, Optional[Dict[str, str]]]:
+def find_best_match(item_name: str, items: List[Dict[str, Any]], word_aliases: List) -> Tuple[
+    int, Optional[Dict[str, str]]]:
     best_score, best_item = 0, None
     common_words = {'arcane', 'prime', 'scene', 'set'}
-    aliases = {'head': 'neuroptics',
-               'taserman': 'volt'}
 
-    item_name = replace_aliases(item_name, aliases)
+    item_name = replace_aliases(item_name, word_aliases)
     item_name = remove_common_words(item_name, common_words)
 
     for item in items:
@@ -126,27 +125,32 @@ class MarketDatabase:
                                    "AND order_type='closed' "
                                    "AND item_id = %s")
     _BASE_ITEMS_QUERY: str = ("SELECT items.id, items.item_name, items.item_type, "
-                              "items.url_name, items.thumb, items.max_rank, item_aliases.alias")
+                              "items.url_name, items.thumb, items.max_rank, GROUP_CONCAT(item_aliases.alias) AS aliases")
 
     _GET_ALL_ITEMS_QUERY: str = (f"{_BASE_ITEMS_QUERY} "
-                                 "FROM items LEFT JOIN item_aliases ON items.id = item_aliases.item_id")
+                                 "FROM items LEFT JOIN item_aliases ON items.id = item_aliases.item_id "
+                                 "GROUP BY items.id")
 
     _GET_ITEMS_IN_SET_QUERY: str = (f"{_BASE_ITEMS_QUERY} "
                                     "FROM items_in_set "
                                     "INNER JOIN items "
                                     "ON items_in_set.item_id = items.id "
                                     "LEFT JOIN item_aliases ON items.id = item_aliases.item_id "
-                                    "WHERE items_in_set.set_id = %s")
+                                    "WHERE items_in_set.set_id = %s "
+                                    "GROUP BY items.id, items.item_name, items.item_type, items.url_name, items.thumb, items.max_rank")
+
+    _GET_ALL_WORD_ALIASES_QUERY: str = "SELECT alias, word FROM word_aliases"
 
     _ADD_ITEM_ALIAS_QUERY: str = "INSERT INTO item_aliases (item_id, alias) VALUES (%s, %s)"
+    _REMOVE_ITEM_ALIAS_QUERY: str = "DELETE FROM item_aliases WHERE item_id=%s AND alias=%s"
 
     _ADD_WORD_ALIAS_QUERY: str = "INSERT INTO word_aliases (word, alias) VALUES (%s, %s)"
 
     _GET_USER_QUERY = "SELECT ingame_name FROM market_users WHERE user_id=%s"
     _UPSERT_USER_QUERY = """
         INSERT INTO market_users (user_id, ingame_name) 
-        VALUES (%s, %s) 
-        ON DUPLICATE KEY UPDATE ingame_name=VALUES(ingame_name)
+        VALUES (%s, %s) AS new
+        ON DUPLICATE KEY UPDATE ingame_name=new.ingame_name
     """
     _INSERT_USERNAME_HISTORY_QUERY = """
         INSERT INTO username_history (user_id, ingame_name, datetime) 
@@ -158,99 +162,158 @@ class MarketDatabase:
         WHERE LOWER(ingame_name) = LOWER(%s)
     """
 
-    def __init__(self, user: str, password: str, host: str, database: str) -> None:
-        self.connection: Connection = pymysql.connect(user=user,
-                                                      password=password,
-                                                      host=host,
-                                                      database=database)
+    _FETCH_ALL_USERS_QUERY = """
+        SELECT user_id, ingame_name FROM market_users
+    """
 
-        self.all_items = self._get_all_items()
+    _GET_PRICE_HISTORY_QUERY = """
+        SELECT datetime, avg_price
+        FROM item_statistics
+        WHERE item_id=%s
+        AND order_type='closed'
+    """
 
-    def _get_all_items(self) -> List[dict]:
-        all_data = self._execute_query(self._GET_ALL_ITEMS_QUERY)
+    _GET_DEMAND_HISTORY_QUERY = """
+        SELECT datetime, volume
+        FROM item_statistics
+        WHERE item_id=%s
+        AND order_type='closed'
+    """
 
-        item_dict = defaultdict(lambda: defaultdict(list))
+    def __init__(self) -> None:
+        self.pool = None
+        self.all_items = None
+
+    @classmethod
+    async def connect(cls, user: str, password: str, host: str, database: str) -> 'MarketDatabase':
+        self = cls()
+        self.pool = await aiomysql.create_pool(user=user, password=password, host=host, db=database)
+
+        self.all_items = await self._get_all_items()
+
+        return self
+
+    async def _execute_query(self, query: str, *params: Union[tuple, List[tuple]], fetch: str = 'all',
+                             commit: bool = False, many: bool = False) -> Union[Tuple, List[Tuple], None]:
+        async with self.pool.acquire() as conn:
+            async with conn.cursor() as cur:
+                if many:
+                    await cur.executemany(query, params[0])
+                else:
+                    await cur.execute(query, params)
+
+                if commit:
+                    await conn.commit()
+
+                if fetch == 'one':
+                    return await cur.fetchone()
+                elif fetch == 'all':
+                    return await cur.fetchall()
+
+    async def _get_all_items(self) -> List[dict]:
+        all_data = await self._execute_query(self._GET_ALL_ITEMS_QUERY, fetch='all')
+
+        all_items: List[dict] = []
         for item_id, item_name, item_type, url_name, thumb, max_rank, alias in all_data:
-            if not item_dict[item_id]['item_data']:
-                item_dict[item_id]['item_data'] = {'id': item_id, 'item_name': item_name, 'item_type': item_type,
-                                                   'url_name': url_name, 'thumb': thumb, 'max_rank': max_rank}
+            aliases = []
             if alias:
-                item_dict[item_id]['aliases'].append(alias)
+                aliases = alias.split(',')
 
-        all_items: List[dict] = [item_data['item_data'] for item_data in item_dict.values()]
+            all_items.append({'id': item_id, 'item_name': item_name, 'item_type': item_type,
+                              'url_name': url_name, 'thumb': thumb, 'max_rank': max_rank, 'aliases': aliases})
 
         return all_items
 
-    def _execute_query(self, query: str, *params) -> Cursor:
-        self.connection.ping(reconnect=True)
-
-        with self.connection.cursor() as cursor:
-            cursor.execute(query, params)
-            return cursor
-
-    async def get_user(self, user: str) -> Optional[MarketUser]:
-        cursor = self._execute_query(self._GET_CORRECT_CASE_QUERY, user)
-        result = cursor.fetchone()
+    async def get_user(self, user: str,
+                       fetch_user_data: bool = True,
+                       fetch_orders: bool = True,
+                       fetch_reviews: bool = True) -> Optional[MarketUser]:
+        result = await self._execute_query(self._GET_CORRECT_CASE_QUERY, user, fetch='one')
 
         if result is None:
             # Username not found, return None
             return None
         else:
             # Return the correct casing
-            return await MarketUser.create(self, result[0], result[1])
+            return await MarketUser.create(self, result[0], result[1],
+                                           fetch_user_data=fetch_user_data,
+                                           fetch_orders=fetch_orders,
+                                           fetch_reviews=fetch_reviews)
 
-    async def get_item(self, item: str) -> Optional[MarketItem]:
-        fuzzy_item = self._get_fuzzy_item(item)
+    async def get_item(self, item: str, fetch_orders: bool = True,
+                       fetch_parts: bool = True, fetch_part_orders: bool = True) -> Optional[MarketItem]:
+        fuzzy_item = await self._get_fuzzy_item(item)
 
         if fuzzy_item is None:
             return None
 
         item_data: List[str] = list(fuzzy_item.values())
 
-        return await MarketItem.create(self, *item_data)
+        return await MarketItem.create(self, *item_data,
+                                       fetch_parts=fetch_parts,
+                                       fetch_orders=fetch_orders,
+                                       fetch_part_orders=fetch_part_orders)
 
-    def get_item_statistics(self, item_id: str) -> Tuple[Tuple[Any, ...], ...]:
-        return self._execute_query(self._GET_ITEM_STATISTICS_QUERY, item_id).fetchall()
+    async def get_item_statistics(self, item_id: str) -> Tuple[Tuple[Any, ...], ...]:
+        return await self._execute_query(self._GET_ITEM_STATISTICS_QUERY, item_id, fetch='all')
 
-    def get_item_volume(self, item_id: str, days: int = 31) -> Tuple[Tuple[Any, ...], ...]:
-        return self._execute_query(self._GET_ITEM_VOLUME_QUERY, days, item_id).fetchall()
+    async def get_item_volume(self, item_id: str, days: int = 31) -> Tuple[Tuple[Any, ...], ...]:
+        return await self._execute_query(self._GET_ITEM_VOLUME_QUERY, days, item_id, fetch='all')
 
-    def close(self) -> None:
-        self.connection.close()
+    async def get_item_price_history(self, item_id: str) -> Dict[str, str]:
+        return dict(await self._execute_query(self._GET_PRICE_HISTORY_QUERY, item_id, fetch='all'))
 
-    def _get_fuzzy_item(self, item_name: str) -> Optional[Dict[str, str]]:
-        best_score, best_item = find_best_match(item_name, self.all_items)
+    async def get_item_demand_history(self, item_id: str) -> Dict[str, str]:
+        return dict(await self._execute_query(self._GET_DEMAND_HISTORY_QUERY, item_id, fetch='all'))
+
+    async def _get_fuzzy_item(self, item_name: str) -> Optional[Dict[str, str]]:
+        best_score, best_item = find_best_match(item_name, self.all_items, await self.get_word_aliases())
 
         return best_item if best_score > 50 else None
 
-    def get_item_parts(self, item_id: str) -> Tuple[Tuple[Any, ...], ...]:
-        return self._execute_query(self._GET_ITEMS_IN_SET_QUERY, item_id).fetchall()
+    async def get_item_parts(self, item_id: str) -> Tuple[Tuple[Any, ...], ...]:
+        return await self._execute_query(self._GET_ITEMS_IN_SET_QUERY, item_id, fetch='all')
 
-    def add_item_alias(self, item_id, alias):
-        return self._execute_query(self._ADD_ITEM_ALIAS_QUERY, item_id, alias)
+    async def get_word_aliases(self) -> Dict[str, str]:
+        return dict(await self._execute_query(self._GET_ALL_WORD_ALIASES_QUERY, fetch='all'))
 
-    def add_word_alias(self, word, alias):
-        return self._execute_query(self._ADD_WORD_ALIAS_QUERY, word, alias)
+    async def add_item_alias(self, item_id, alias):
+        await self._execute_query(self._ADD_ITEM_ALIAS_QUERY, item_id, alias, commit=True)
+        self.all_items = await self._get_all_items()
 
-    def update_usernames(self, data: dict) -> None:
-        with ThreadPoolExecutor() as executor:
-            executor.submit(self._background_update_usernames, data)
+    async def remove_item_alias(self, item_id, alias):
+        await self._execute_query(self._REMOVE_ITEM_ALIAS_QUERY, item_id, alias, commit=True)
+        self.all_items = self._get_all_items()
 
-    def _background_update_usernames(self, data: dict) -> None:
+    async def add_word_alias(self, word, alias):
+        await self._execute_query(self._ADD_WORD_ALIAS_QUERY, word, alias, commit=True)
+
+    async def update_usernames(self, data: dict) -> None:
+        # Fetch all user data first
+        user_data = dict(await self._execute_query(self._FETCH_ALL_USERS_QUERY, fetch='all'))
+
+        # Prepare batch queries
+        update_queries = []
+        history_queries = []
+        now = datetime.now()
         for user_id, new_ingame_name in data.items():
-            result = self._execute_query(self._GET_USER_QUERY, user_id).fetchone()
+            current_ingame_name = user_data.get(user_id)
 
             # If the user doesn't exist or username is different,
             # update the user's username in `market_users` and add a record in `username_history`
-            if result is None or new_ingame_name != result[0]:
+            if current_ingame_name is None or new_ingame_name != current_ingame_name:
                 logger.info(f"Updating username for user {user_id} to {new_ingame_name}")
-                self._execute_query(self._UPSERT_USER_QUERY, user_id, new_ingame_name)
 
-                now = datetime.now()
-                self._execute_query(self._INSERT_USERNAME_HISTORY_QUERY, user_id, new_ingame_name, now)
+                update_queries.append((user_id, new_ingame_name))
+                history_queries.append((user_id, new_ingame_name, now))
+
+        # Execute the queries
+        if update_queries:
+            await self._execute_query(self._UPSERT_USER_QUERY, update_queries, commit=True, many=True)
+
+        if history_queries:
+            await self._execute_query(self._INSERT_USERNAME_HISTORY_QUERY, history_queries, commit=True, many=True)
 
-            # Commit your changes
-            self.connection.commit()
 
 
 def require_orders():
@@ -285,7 +348,7 @@ class MarketItem:
 
     def __init__(self, database: MarketDatabase,
                  item_id: str, item_name: str, item_type: str, item_url_name: str, thumb: str, max_rank: str,
-                 fetch_orders: bool = True, fetch_parts: bool = True, fetch_part_orders: bool = True) -> None:
+                 aliases: List, fetch_orders: bool = True, fetch_parts: bool = True, fetch_part_orders: bool = True) -> None:
         self.database: MarketDatabase = database
         self.item_id: str = item_id
         self.item_name: str = item_name
@@ -293,25 +356,52 @@ class MarketItem:
         self.item_url_name: str = item_url_name
         self.thumb: str = thumb
         self.max_rank: str = max_rank
+        self.aliases: List = aliases
         self.thumb_url: str = f"{MarketItem.asset_url}/{self.thumb}"
         self.item_url: str = f"{MarketItem.base_url}/{self.item_url_name}"
         self.orders: Dict[str, List[Dict[str, Union[str, int]]]] = {'buy': [], 'sell': []}
         self.parts: List[MarketItem] = []
+        self.part_orders_fetched: bool = False
+        self.part_price_history_fetched: bool = False
+        self.part_demand_history_fetched: bool = False
+        self.price_history: Dict[str, str] = {}
+        self.demand_history: Dict[str, str] = {}
+        self.users: Dict[str, str] = {}
 
     @classmethod
     async def create(cls, database: MarketDatabase, item_id: str, item_name: str, item_type: str,
-                     item_url_name: str, thumb: str, max_rank: str, fetch_orders: bool = True,
-                     fetch_parts: bool = True, fetch_part_orders: bool = True):
-        obj = cls(database, item_id, item_name, item_type, item_url_name, thumb, max_rank)
+                     item_url_name: str, thumb: str, max_rank: str, aliases: List, fetch_orders: bool = True,
+                     fetch_parts: bool = True, fetch_part_orders: bool = True,
+                     fetch_price_history: bool = True, fetch_demand_history: bool = True):
+        obj = cls(database, item_id, item_name, item_type, item_url_name, thumb, max_rank, aliases)
 
+        tasks = []
         if fetch_orders:
-            await obj.get_orders()
+            tasks.append(obj.get_orders())
 
         if fetch_parts:
-            obj.get_parts()
+            tasks.append(obj.get_parts())
 
         if fetch_part_orders:
-            await obj.get_part_orders()
+            tasks += obj.get_part_orders_tasks()
+
+            if fetch_price_history:
+                for part in obj.parts:
+                    if part is not None:
+                        await part.get_price_history()
+
+            if fetch_demand_history:
+                for part in obj.parts:
+                    if part is not None:
+                        await part.get_demand_history()
+
+        if fetch_price_history:
+            await obj.get_price_history()
+
+        if fetch_demand_history:
+            await obj.get_demand_history()
+
+        await asyncio.gather(*tasks)
 
         return obj
 
@@ -329,8 +419,23 @@ class MarketItem:
 
         return filters, mode
 
-    def add_alias(self, alias: str) -> None:
-        self.database.add_item_alias(self.item_id, alias)
+    def get_part_orders_tasks(self) -> list[Coroutine[Any, Any, None]]:
+        tasks = [part.get_orders() for part in self.parts if part is not None]
+        self.part_orders_fetched = True
+
+        return tasks
+
+    async def get_price_history(self) -> None:
+        self.price_history = await self.database.get_item_price_history(self.item_id)
+
+    async def get_demand_history(self) -> None:
+        self.demand_history = await self.database.get_item_demand_history(self.item_id)
+
+    async def add_alias(self, alias: str) -> None:
+        await self.database.add_item_alias(self.item_id, alias)
+
+    async def remove_alias(self, alias: str) -> None:
+        await self.database.remove_item_alias(self.item_id, alias)
 
     def filter_orders(self,
                       order_type: str = 'sell',
@@ -389,7 +494,7 @@ class MarketItem:
 
         return filtered_orders[:num_orders]
 
-    def parse_orders(self, orders: List[Dict[str, Any]]) -> None:
+    async def parse_orders(self, orders: List[Dict[str, Any]]) -> None:
         self.orders: Dict[str, List[Dict[str, Union[str, int]]]] = {'buy': [], 'sell': []}
         users = {}
 
@@ -417,21 +522,17 @@ class MarketItem:
         for key, reverse in [('sell', False), ('buy', True)]:
             self.orders[key].sort(key=lambda x: (x['price'], x['last_update']), reverse=reverse)
 
-        self.database.update_usernames(users)
+        self.users.update(users)
 
     async def get_orders(self) -> None:
         orders = await fetch_wfm_data(f"{self.base_api_url}/items/{self.item_url_name}/orders")
         if orders is None:
             return
 
-        self.parse_orders(orders['payload']['orders'])
+        await self.parse_orders(orders['payload']['orders'])
 
-    def get_parts(self) -> None:
-        self.parts = [MarketItem(self.database, *item) for item in self.database.get_item_parts(self.item_id)]
-
-    async def get_part_orders(self) -> None:
-        tasks = [part.get_orders() for part in self.parts]
-        await asyncio.gather(*tasks)
+    async def get_parts(self) -> None:
+        self.parts = [MarketItem(self.database, *item) for item in await self.database.get_item_parts(self.item_id)]
 
 
 class MarketUser:
@@ -439,7 +540,7 @@ class MarketUser:
     base_url: str = "https://warframe.market/profile"
     asset_url: str = "https://warframe.market/static/assets"
 
-    def __init__(self, database: MarketDatabase, user_id: int, username: str):
+    def __init__(self, database: MarketDatabase, user_id: str, username: str):
         self.database = database
         self.user_id = user_id
         self.username = username
@@ -456,21 +557,24 @@ class MarketUser:
         self.status = None
         self.region = None
         self.orders: Dict[str, List[Dict[str, Union[str, int]]]] = {'buy': [], 'sell': []}
-        self.reviews: Dict[str, List[Dict[str, Union[str, int]]]] = []
+        self.reviews: List[str] = []
 
     @classmethod
     async def create(cls, database: MarketDatabase, user_id: str, username: str,
                      fetch_user_data: bool = True, fetch_orders: bool = True, fetch_reviews: bool = True):
         obj = cls(database, user_id, username)
 
+        tasks = []
         if fetch_user_data:
-            await obj.fetch_user_data()
+            tasks.append(obj.fetch_user_data())
 
         if fetch_orders:
-            await obj.fetch_orders()
+            tasks.append(obj.fetch_orders())
 
         if fetch_reviews:
-            await obj.fetch_reviews()
+            tasks.append(obj.fetch_reviews())
+
+        await asyncio.gather(*tasks)
 
         return obj
 
